# coding:utf-8
#
# unike/module/__init__.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 28, 2023
# 
# 该头文件定义了 module 接口.

"""模块部分，包含模型和损失函数。"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .BaseModule import BaseModule

__init__ = {
	'BaseModule'
}